package week02

import Eloadas._

import scala.annotation.tailrec
import scala.reflect.internal.pickling.PickleFormat

object Gyakorlat extends App {
  /**
   * Implementáljuk a Fibonacci sorozatot, Fib( 0 ) = Fib ( 1 ) = 1, Fib ( n ) = Fib( n - 1 ) + Fib( n - 2 ), tail rekurzívan!
   */
  def fib(n : Int): Int = {
    def fib( n: Int, prev: Int, current: Int): Int =
      if(n <= 1) current else fib(n-1, current, prev+current)
    fib(n,1,1)
  }

  /**
   * Írjunk algoritmust, ami visszaadja az input IntListánk hosszát!
   */
  def length( list : IntLista ) : Int = list match {
    case UresLista => 0
    case NemuresLista(_, tail) => 1 + length(tail)
  }

  /**
   * Írjunk algoritmust, ami megfordítja az inputként kapott IntListánkat!
   * Hint: készítsünk (tail recursive) kétváltozós belső függvényt hozzá.
   */
  def reverse(list: IntLista ): IntLista = {
    def reverse(list: IntLista, result: IntLista): IntLista = list match {
      case UresLista => result
      case NemuresLista(head, tail) => reverse(tail, NemuresLista(head,result))
    }
    reverse(list, UresLista)
  }

  /**
   * Írjunk függvényt, ami "pretty print"eli az IntListát!
   * Pl: üres lista (), egyelemű lista: (4), hatelemű lista: (1,4,2,8,5,7)
   * (zárójelekkel együtt! hint: egy belső (tailrec?) függvénnyel építsük fel
   * a zárójel nélküli részt, majd rakjuk köré a zárójeleket.)
   */
  def prettyPrint( list : IntLista ) : String = {
    def prettyPrintInner(list: IntLista): String = list match {
      case UresLista => ""
      case NemuresLista(head, UresLista) => head.toString
      case NemuresLista(head, tail) => head +"," +prettyPrintInner(tail)
    }
    "(" + prettyPrintInner(list) + ")"
  }

  /**
   * Írjunk algoritmust, ami szétvág egy listát két listára, melyek hossza közt a különbség nem lehet nagyobb 1-nél
   * Pl. az (1,4,2,8,5,7) listából az ((1,2,5),(4,8,7)) párt,
   * az (1,2,3,4,5) listából az ((1,3,5),(2,4)) párt képezheti.
   * Párt egyszerűen így adunk vissza:
   */
  def rendez( x : Int, y : Int ) : (Int,Int) = if( x < y ) (x,y) else (y,x)

  def split( list : IntLista ) : (IntLista, IntLista) = list match {
    case UresLista => (UresLista,UresLista)
    case NemuresLista(_, UresLista) => (list, UresLista)
    case NemuresLista(head1, NemuresLista(head2, tail)) => {
      val (tail1,tail2) = split(tail)
      (NemuresLista(head1,tail1), NemuresLista(head2,tail2))
    }
  }
  /**
   * Írjunk algoritmust, ami összemergel két rendezett IntListát egy rendezett IntListába!
   * pl. a (2,3,6) és a (1,2,4) listák mergelése a (1,2,2,3,4,6) lista legyen.
   * Ellenőrizni nem kell, hogy a listák valóban rendezettek-e.
   * Hint: párra is lehet matchelni.
   */
  def merge( left : IntLista, right : IntLista ): IntLista = (left,right) match {
    case (UresLista,_) => right
    case (_,UresLista) => left
    case (NemuresLista(head1,tail1), NemuresLista(head2,_)) if (head1<head2) => NemuresLista( head1, merge(tail1,right))
    case (_,NemuresLista(head2,tail2)) => NemuresLista(head2, merge(left,tail2))
  }

  /**
   * Írjunk algoritmust, ami mergesortol egy listát:
   * - szétvágja két majdnem egyforma hosszú részlistára,
   * - rendezi a részeket rekurzívan,
   * - majd összemergeli a két rendezett részt!
   */
  def mergesort( list : IntLista ) : IntLista =  list match {
    case UresLista => UresLista
    case NemuresLista(_,UresLista) => list
    case _ => {
      val (list1, list2) = split(list)
      merge(mergesort(list1), mergesort(list2))
    }
  }

  /**
   * kalkulusból tanultuk, hogy a határozott integrál egy az [a,b] intervallumonon
   * az f függvény alatti terület. Írjunk függvényt, ami paraméterként megkapja
   * az f függvényt, az a és b végpontokat és visszaadja az a-b közti trapéz alatti területet!
   *
   */
  def trapezSzabaly( f : Double => Double , a : Double, b : Double ) : Double = (b-a) * (f(a)+f(b)) / 2.0

  /**
   * A "trapéz módszer" annyiban más a trapéz szabálytól, hogy az a,b intervallumot
   * felosztja n egyforma részre és külön-külön alkalmazza a kicsi intervallumokra a
   * trapéz szabályt. Majd persze összeadja az eredményeket.
   * Implementáld a trapéz módszert is! n az intervallumok száma.
   * Ha n <= 0, adjon vissza Double.NaN -t.
   */
  def trapezModszer( f : Double => Double, a : Double, b : Double, n : Int ) : Double  =
    if (n <= 0) Double.NaN
    else {
      @tailrec
      def trapezModszer(left: Double, right: Double, segment: Double, n: Int, acc: Double ):Double =
        if(n==0) acc else trapezModszer(left+segment, right, segment, n-1, acc+trapezSzabaly(f,left,left+segment) )
      trapezModszer(a,b,(b-a)/n,n,0.0)
    }

  /**
   * A szelő módszer: f, a, b. Biztos, hogy f(a) < 0 és f(b) > 0 vagy fordítva (tehát ellentétes az előjelük).
   * Zérushelyet keresünk a következő módon:
   * i) vesszük a felezőpontot, (a+b)/2, nézzük itt a függvényértéket
   * ha ez abszolútértékben kisebb, mint egy paraméterként kapott epszilon, akkor visszaadjuk a felezőpontot
   * ha nem, akkor ha pozitív, akkor az a,b közül a negatívval, ha negatív, akkor a pozitívval hívjuk rekurzívan
   * az algoritmust.
   *
   * implementáljuk!
   */
  def hurModszer( f : Double => Double, a : Double, b : Double, epszilon : Double ) : Double = {
    val mid = (a+b)/2.0
    val fmid = f(mid)
    if (Math.abs(fmid) < epszilon) mid
    else if(fmid * f(a) < 0) hurModszer(f,a,mid,epszilon)
    else hurModszer(f,mid,b,epszilon)
  }

  /**
   * Ha ezzel is kész vagyunk, akkor implementáljunk bináris keresőfát, mely támogatja
   * - üres fa létrehozását
   * - fába egy érték beszúrását, a keresőfa tulajdonság fenntartása mellett, visszaadva egy új fát,
   * - fában egy érték keresését , Boolean visszatérési értékkel
   * - két fa unióját, egy új faként
   * - fa méretének visszaadását!
   *
   * példákat a táblán keressünk -- tervezzük meg hozzá az adatszerkezeetet is!
   */
}
