package week09

/**
 * A Heap, azaz Kupac adatszerkezet hasonlít egy bináris keresőfára annyiban, hogy kulcsokat használ, amik rendezettek;
 * gyorsan támogatja a *legkisebb* kulcsú elem lekérését, ezért prioritási sorok implementálásakor használnak.
 *
 * Itt egy fában egy (key,value) párt tárolunk, minden fapontnak van két gyereke (bármelyik lehet üres fa is),
 * a kupac tulajdonság pedig: minden fapont kulcsa kisebb vagy egyenlő, mint a gyerekeinek a kulcsa.
 * (Így a legkisebb kulcs mindig a fa gyökerében található).
 *
 * Minden csúcsnak legyen egy size *adattagja*, amiben a benne levő (prioritás, érték) párok számát tároljuk.
 * (Üres fáé nulla)
 *
 * Beszúráskor egy ötlet:
 * - ha egy (prioritás,érték) párt beszúrunk egy üres fába, akkor jöjjünk vissza egy egypontú fával;
 * - ha egy nemüres fába, és a beszúrandó prioritás kisebb, mint az aktuális gyökércsúcs prioritása, akkor ez legyen
 *   az új gyökér, és a régi gyökéradatokat szúrjuk be a kisebb méretű gyerekbe;
 * - ha egy nemüres fába, a gyökérprioritásnál nagyobb-egyenlő a beszúrandó prioritás, akkor szúrjuk be a
 *   kisebb méretű gyerekbe a csúcsot
 * (mindezt persze funkcionálisan, az eredeti immutable adatszerkezetet nem bántva)
 *
 * Pophoz egy ötlet:
 * - üres fán dobjon UnsupportedOperationException -t
 * - ha a fának amiből popolunk van üres gyereke, akkor az egyetlen megmaradó gyereket érdemes visszaadni,
 * - ha mindkét gyereke nemüres, akkor popoljuk ki abból a minimumot, amelyikben kisebb a minimális kulcs,
 *   és ez legyen az új gyökércsúcs
 */

/**
 * Implementáljuk a ???-es metódusok mindegyikét! Némelyiket lehet, hogy a traitben érdemes implementálni,
 * némelyeket a két leszármazott osztályban.
 */
trait PriorityMap[Key <: Ordered[Key], Value] {
  def isEmpty: Boolean = ???  //adja vissza, hogy üres-e a prisor!
  def minKey: Key = ??? //adja vissza a minimális prioritást! Ha nincs elem egyáltalán, dobjon UnsupportedOperationException("EmptyPriorityMap.minKey") -t.
  def minEntry: (Key,Value) = ???  //adja vissza az egyik minimális prioritású entryt! Ha nincs, dobjon UnsupportedOperationException("EmptyPriorityMap.minEntry") -t.
  def size: Int = ??? //adja vissza a prisor méretét (hány (prioritás,érték) pár van benne
  def +(key: Key, value: Value): PriorityMap[Key,Value] = ??? //szúrjon be egy új (prioritás,érték) párt! Azonos prioritású érték lehet több is az adatszerkezetben.
  def pop: PriorityMap[Key,Value] = ??? //adjon vissza egy olyan prisort, mely a minimális prioritásút (amit a minEntry visszaad) nem tartalmazza az entryk közül
  def ++(that: PriorityMap[Key,Value]): PriorityMap[Key,Value] = ??? //adja vissza a két prisor unióját: olyan prisort, ami tartalmazza az összes entryt, ami a két prisorban együtt van.
  def toList: List[Value] = ??? //adja vissza a prisor elemeinek az érték részeit egy olyan listában, melybe prioritás szerint növekvőbe rendezettek.
  def foreach[U](f: (Key,Value) => U) : Unit = ??? //értékelje ki az f(key,value) kifejezést a prisor összes (key,value) entryjén.
  def map[Key2 <: Ordered[Key2], Value2](f: (Key,Value) => (Key2,Value2)): PriorityMap[Key2,Value2] = ??? //mappelje le az összes (prioritás, érték) párt f-fel és az eredményekből építsen egy új prisort.
  def filter(p: (Key,Value) => Boolean): PriorityMap[Key,Value] = ??? //szűrje le, ahogy a filterrel szoktuk.
}

//Az üres prisor case classa.
case class EmptyPriorityMap[Key <: Ordered[Key], Value]() extends PriorityMap[Key, Value] {
}

//A nemüres prisor case classa: key,value a gyökérelemben tárolt (prioritás,érték) pár, left és right pedig a két gyerek prisor.
case class NonemptyPriorityMap[Key <: Ordered[Key], Value]
  (left: PriorityMap[Key,Value],
   key: Key,
   value: Value,
   right: PriorityMap[Key,Value])
  extends PriorityMap[Key, Value]{
}

/**
 * Feladat még a week09.PracticeTest -ben is implementálni teszteket, amikkel a saját implementációnkat mérjük le!
 */

object Practice {

}
