package week02

import org.scalatest.concurrent.{Signaler, TimeLimitedTests}
import org.scalatest.flatspec.AnyFlatSpec
import org.scalatest.time.{Millis, Span}

import scala.annotation.tailrec
import scala.util.Random

/**
 * tesztek a második gyakhoz:
 */
class GyakorlatTeszt extends AnyFlatSpec with TimeLimitedTests{
  val timeLimit : Span = Span( 1000, Millis )
  override val defaultTestSignaler: Signaler = new Signaler{
    override def apply(testThread: Thread): Unit = {
      println("Ez a teszt túl sokáig fut.")
      testThread.stop() } //unsafe, never használd.
  }

  @tailrec
  private def loop : Boolean = loop

  "fibonacci" should "be 0 -> 1" in assert( Gyakorlat.fib(0) == 1 )
  it should "be 1 -> 1" in assert( Gyakorlat.fib(1) == 1 )
  it should "be 2 -> 2" in assert( Gyakorlat.fib(2) == 2 )
  it should "be 5 -> 8" in assert( Gyakorlat.fib(5) == 8 )
  it should "be tail recursive" in assert( Gyakorlat.fib(20000) == 0 || true )

  def generate( list : List[Int] , intList : IntLista ) : (List[ Int ], IntLista) = Random.nextInt(5) match {
    case 0 => (list,intList)
    case _ => val head = Random.nextInt(20); generate( head :: list , NemuresLista( head, intList ))
  }
  def generateList = generate( Nil , UresLista )
  def sameList( list : List[Int], lista : IntLista ) : Boolean = (list,lista) match {
    case (Nil,UresLista) => true
    case (head :: tail, NemuresLista( head2, tail2 )) if ( head == head2 ) => sameList(tail, tail2)
    case _ => false
  }
  "length" should "return the length of the list" in{
    for( i <- 1 to 1000 ) {
      val (list,lista) = generateList
      assert( list.length == Gyakorlat.length( lista ) , "length of " + list + " became " + Gyakorlat.length( lista ) )
    }
  }

  "reverse" should "return empty list on empty list"  in assert ( Gyakorlat.reverse( UresLista  ) == UresLista )
  it should "coincide with the library reverse implementation" in {
    for( i <- 1 to 1000 ) {
      val (list,lista) = generateList
      val revList = list.reverse
      assert( sameList( revList, Gyakorlat.reverse(lista)), "error on " + list + ", result is " + Gyakorlat.reverse(lista))
    }
  }
  it should "work for long lists" in {
    def longList( n : Int, list : IntLista ) : IntLista = if( n == 0 ) list else longList( n-1, NemuresLista(n, list ))
    assert( Gyakorlat.reverse( longList(20000, UresLista) ) != UresLista )
  }

  def toLista( list : List[Int] ) : IntLista = list match {
    case Nil => UresLista
    case head :: tail => NemuresLista( head, toLista(tail))
  }
  def toList( lista : IntLista ) : List[Int] = lista match {
    case UresLista => Nil
    case NemuresLista(head, tail) => head :: toList( tail )
  }
  "prettyPrint" should "print the lists nicely" in {
    for( i <- 1 to 100 ) {
      val (list, intList) = generateList
      assert( Gyakorlat.prettyPrint( intList ) == list.mkString("(",",",")" ), "List " + intList + " is printed as " + Gyakorlat.prettyPrint( intList ) )
    }
  }

  "merge" should "merge two ordered lists" in {
    for( i <- 1 to 100 ) {
      val list1 = generateList._1.sorted
      val list2 = generateList._1.sorted
      val merged = Gyakorlat.merge(toLista(list1), toLista(list2))
      assert(sameList((list1 ::: list2).sorted, merged), "merge of " + list1 + " and " + list2 + " became " + merged)
    }
  }

  "split" should "split a list into two almost equally large parts" in {
    for( i <- 1 to 100 ) {
      val (list, intList) = generateList
      val (left, right) = Gyakorlat.split( intList )
      assert( Math.abs( Gyakorlat.length( left ) - Gyakorlat.length( right ) ) <= 1 , "split of " + list + " became " + left + " vs " + right )
      val leftList = toList( left )
      val rightList = toList( right )
      assert( list.sorted == (leftList ::: rightList).sorted , "split of " + list + " became " + left + " vs " + right )
    }
  }

  "mergesort" should "sort its input" in {
    for( i <- 1 to 1000) {
      val (list, intList) = generateList
      val sortedLista = Gyakorlat.mergesort( intList )
      assert( sameList( list.sorted, sortedLista ), "mergesort returns " + sortedLista + " for the input " + list)
    }
  }

  "trapezszabaly" should "return 0 for sin(x) in 0..PI" in assert( Math.abs( Gyakorlat.trapezSzabaly( Math.sin , 0.0, Math.PI ) ) < 1e-10 )
  it should "return 1/2 for x in 0..1" in assert(  Gyakorlat.trapezSzabaly( { x => x }, 0.0, 1.0 ) == 0.5 )
  it should "return 0 for 2x in -1..1" in assert(  Gyakorlat.trapezSzabaly( { x => 2*x }, -1.0, 1.0 ) == 0.0 )
  it should "return 2 for x^2 in -1..1" in assert( Gyakorlat.trapezSzabaly( {x => x*x}, -1.0, 1.0 ) == 2.0 )

  "trapezmodszer" should "return NaN if n <= 0 " in assert( Gyakorlat.trapezModszer( { x => x }, 0.0, 0.0 , 0 ).isNaN )
  it should "return 2 for x^2 in -1..1 with n=1" in assert( Gyakorlat.trapezModszer( { x => x*x }, -1.0, 1.0, 1 ) == 2.0 )
  it should "return 1 for x^2 in -1..1 with n=2" in assert( Gyakorlat.trapezModszer( { x => x*x }, -1.0, 1.0, 2 ) == 1.0 )
  it should "return ~0.68 for x^2 in -1..1 with n=10" in assert( Math.abs( Gyakorlat.trapezModszer( {x=>x*x}, -1.0, 1.0, 10 ) - 0.68) < 1e-10 )
  it should "be tail recursive and should return 1.0 for sin(x) in 0..PI with n = 50000" in assert( Math.abs(Gyakorlat.trapezModszer( Math.sin, 0.0, Math.PI / 2.0 , 50000 ) -1.0)<1e-10 )

  "hurmodszer" should "find a root of x^2 - 2" in {
    val f = week01.Gyakorlat.createMasodfoku(1.0, 0.0, -2.0)
    val root = Gyakorlat.hurModszer( f, 0.0, 2.0, 1e-10 )
    assert( Math.abs( f(root) ) < 1e-10, "x^2 - 2 is not close enough to 0.0 at x="+root )
  }

}
