package week06

import java.util.NoSuchElementException

import org.scalatest.concurrent.{Signaler, TimeLimitedTests}
import org.scalatest.flatspec.AnyFlatSpec
import org.scalatest.time.{Millis, Span}

import scala.util.Random
import week06.MapIntT

class PracticeTest extends AnyFlatSpec with TimeLimitedTests {
  val timeLimit: Span = Span(1000, Millis)
  override val defaultTestSignaler: Signaler = new Signaler {
    override def apply(testThread: Thread): Unit = {
      println("Ez a teszt túl sokáig fut.")
      testThread.stop()
    } //unsafe, never használd.
  }

  "contains" should "return false for the empty map" in {
    assert(!EmptyMap[String]().asInstanceOf[MapIntT[String]].contains(0))
    assert(!EmptyMap[Double]().asInstanceOf[MapIntT[String]].contains(1))
  }
  it should "return true or false correctly" in {
    for (test <- gettests) assert(test._1.contains(test._2) == (test._3 != None))
  }

  val empty = EmptyMap[String]()
  val one = NonemptyMap(empty, 1, "Egy", empty)
  val twoone = NonemptyMap(one, 2, "Kettő", empty)
  val onetwothree = NonemptyMap(one, 2, "Kettő", NonemptyMap(empty, 3, "Három", empty))

  "size" should "return 0 for the empty map" in {
    assert(empty.size == 0)
  }
  it should "return the number of nodes in the entry set" in {
    assert(one.size == 1, "teszteset: " + one)
    assert(twoone.size == 2, "teszteset: " + twoone)
    assert(onetwothree.size == 3, "teszteset: " + onetwothree)
  }
  "depth" should "return 0 for the empty map" in {
    assert(empty.depth == 0)
  }
  it should "return the length of the longest root-leaf path" in {
    assert(one.depth == 1, "teszteset: " + one)
    assert(twoone.depth == 2, "teszteset: " + twoone)
    assert(onetwothree.depth == 2, "teszteset: " + onetwothree)
  }
  "isEmpty" should "return true for the empty map" in {
    assert(empty.asInstanceOf[MapIntT[String]].isEmpty)
  }
  it should "return false for a nonempty map" in {
    for (test <- List[MapIntT[String]](one, twoone, onetwothree)) assert(!test.isEmpty, "teszteset: " + test)
  }
  "get" should "return None when called on the empty map" in {
    assert(empty.get(100) == None)
  }
  val gettests = List(
    (one, 1, Some("Egy")), (one, 0, None),
    (twoone, 1, Some("Egy")), (twoone, 2, Some("Kettő")), (twoone, 4, None), (twoone, 0, None),
    (onetwothree, 1, Some("Egy")), (onetwothree, 2, Some("Kettő")), (onetwothree, 3, Some("Három")), (onetwothree, 0, None), (onetwothree, 5, None))
  it should "return an Option correctly" in {
    for (test <- gettests) assert(test._1.get(test._2) == test._3)
  }
  "getOrElse" should "return a value or the default correctly" in {
    assert(empty.getOrElse(0, "sanyi") == "sanyi", "empty.getOrElse(0,\"sanyi\")")
    for (test <- gettests) assert(test._1.getOrElse(test._2, "sanyi") == test._3.getOrElse("sanyi"), "teszteset: " + test._1 + ".getOrElse(" + test._2 + ",\"sanyi\")")
  }
  "remove" should "not modify the empty map" in {
    assert(empty.remove(0) == empty)
  }
  it should "handle removal correctly" in {
    for (i <- 0 to 4; j <- 0 to 4) {
      val removedMap = onetwothree.asInstanceOf[MapIntT[String]].remove(i)
      assert(removedMap.contains(j) == (j >= 1 && j <= 3 && i != j), "teszteset: " + onetwothree + " remove(" + i + ").contains(" + j + ") returned " + removedMap.contains(j))
    }
  }
  "maxKey" should "throw an UnsupportedOperationException when called on the empty map" in {
    try {
      empty.maxKey; assert(false)
    } catch {
      case uoe: UnsupportedOperationException => ();
      case _: Throwable => assert(false)
    }
  }
  it should "return the largest key" in {
    val maxTest = List((one, 1), (twoone, 2), (onetwothree, 3))
    for (test <- maxTest) assert(test._1.maxKey == test._2, "teszteset: " + test._1)
  }
  "updated" should "give back a one-element tree when called on the empty map" in {
    assert(empty.updated(1, "Egy") == one, "elvárt: " + one + ", eredmény: " + empty.updated(1, "Egy"))
  }
  it should "insert or modify a nonempty map correctly" in {
    for (
      i <- 0 to 4;
      j <- 0 to 4
    ) {
      val up = onetwothree.asInstanceOf[MapIntT[String]].updated(i, "check")
      val op = up.get(j)
      if (i == j) assert(op == Some("check"), "teszteset: " + onetwothree + ".updated(" + i + ",\"check\").get(" + j + ")")
      else assert(onetwothree.get(j) == op, "teszteset: " + onetwothree + ".updated(" + i + ",\"check\").get(" + j + ")")
    }
  }
  "map" should "return an empty Vector when called on the empty map" in {
    assert(empty.map(x => x) == Vector())
  }
  it should "compute the values of the function in increasing order with respect to the keys" in {
    assert(onetwothree.map(entry => entry._2) == Vector("Egy", "Kettő", "Három"), "teszteset: " + onetwothree + ".map( entry => entry._2 )")
  }
  "foldLeft" should "return the initial value when called on the empty map" in {
    assert(empty.foldLeft("") { (str, entry) => str + "x" } == "")
  }
  it should "fold correctly in an increasing order with respect to the keys" in {
    assert(onetwothree.foldLeft(10)((acc, entry) => 2 * acc + entry._1 + entry._2.length) == 118, "teszteset: " + onetwothree + ".foldLeft(10)( (acc,entry)=>2*acc+entry._1+entry._2.length)")
  }
  "filter" should "return the empty map back when called on it" in {
    assert(empty.filter(tuple => true) == empty)
  }
  "foreach" should "make an inorder traversal" in {
    var s = ""
    onetwothree.foreach(entry => s = s + entry._2)
    assert(s == "EgyKettőHárom", "call: " + onetwothree + ".foreach( entry => print(entry._2) ), elvárt: EgyKettőHárom, eredmény: " + s)
  }
  "++" should "work for union with the empty map" in {
    val res = empty ++ onetwothree
    for( i <- 0 to 4 ) assert(
      res.get(i) == onetwothree.get(i),
      "(MapIntT[String]() ++ " + onetwothree + ").get(" + i + ")"
    )
    val res2 = onetwothree ++ empty
    println(res2)
    for( i <- 0 to 4 ) assert(
      onetwothree.get(i) == res2.get(i),
      "(" + onetwothree + " ++ MapIntT[String]() ).get(" + i + ")"
    )
  }
  "++" should "keep the keys of the right-hand side map" in {
    val threefour = NonemptyMap(empty,3,"Three",NonemptyMap(empty,4,"Four",empty))
    val union = onetwothree ++ threefour
    for( test <- List( (1,"Egy"), (2,"Kettő"), (3,"Three"), (4,"Four")) ) assert(union.get(test._1) == Some(test._2), "teszteset: (" + onetwothree + " ++ " + threefour + ").get(" + test._1 + ")")
    assert( union.get(0) == None,"teszteset: (" + onetwothree + " ++ " + threefour + ").get(0)" )
    assert( union.get(5) == None,"teszteset: (" + onetwothree + " ++ " + threefour + ").get(5)" )
  }

}

