package week09

import org.scalatest.concurrent.{Signaler, TimeLimitedTests}
import org.scalatest.flatspec.AnyFlatSpec
import org.scalatest.time.{Millis, Span}

class PracticeTest extends AnyFlatSpec with TimeLimitedTests {
  val timeLimit: Span = Span(1000, Millis)
  override val defaultTestSignaler: Signaler = new Signaler {
    override def apply(testThread: Thread): Unit = {
      println("Ez a teszt túl sokáig fut.")
      testThread.stop()
    } //unsafe, never használd.
  }

  /**
   * A tesztekben a kulcs típusa a MyKeyType lesz, aminek egyetlen string adattagja van, és eszerint rendez
   */
  case class MyKeyType(key: String) extends Ordered[MyKeyType] {
    /**
     * Implementáljunk összehasonlító metódust, mely elsősorban a key stringek hossza alapján, másodsorban
     * kisbetűsített változat szerinti ábécében rendez!
     * tehát pl. MyKeyType("abc").compare(MyKeyType("de")) pozitív legyen, hiszen "de" rövidebb, mint "abc".
     * A compare metódus visszatérési értéke
     * - negatív, ha this kisebb, mint that,
     * - pozitív, ha that kisebb, mint this,
     * - nulla, ha a kettő egyenlő.
     */
    override def compare(that: MyKeyType): Int = ???
  }

  /**
   * A tesztekben építsünk PriorityMap[MyKeyType,String] -eket, azaz a prioritás és a tárolt érték is legyen String.
   * Ezekhez írjunk egy random String generátor metódust, ami random hosszú stringeket generál, melyek az angol
   * ábécé kisbetűit vagy nagybetűit tartalmazzák!
   */
  def generateLowercaseString: String = ???
  def generateUppercaseString: String = ???

  /**
   * Írjunk egy random (MyKeyType,String) párokat generáló metódust, ami egy MyKeyType(key)-t generál véletlen
   * uppercase prioritással kulcsnak és egy lowercase random stringet értéknek!
   */
  def generateRandomEntry: (MyKeyType,String) = ???

  /**
   * Írjunk tesztet, ami azt ellenőrzi, hogy az üres prisornak 0-e a mérete!
   */
  "EmptyPriorityMap.size()" should "be 0" in ???

  /**
   * Írjunk tesztet, ami azt ellenőrzi, hogy az üres prisor üres-e (isEmpty == true)!
   */
  "EmptyPriotityMap.isEmpty" should "be true" in ???

  /**
   * Írjunk tesztet, ami azt ellenőrzi, hogy egy random generált (prioritás,érték) lista elemeit ha beszúrjuk
   * egy üres prisorba, akkor az eredmény prisor mérete megegyezik-e a lista méretével!
   * Ilyen random generált tesztből futtassunk százat.
   */
  "Insert" should "produce a PriorityMap of correct size" in ???

  /**
   * Írjunk tesztet, ami azt ellenőrzi, hogy ha egy random entryt beszúrunk egy üres prisorba,
   * akkor a minkey ennek adja-e vissza a kulcsát! Ilyen random tesztből futtassunk százat.
   */
  "Insert+pop" should "give back the empty priority queue" in ???

  /**
   * Írjunk egy helper metódust, ami egy List[(MyKeyType,String)]-nek visszaadja azt az elemét,
   * amelyik a legkisebb MyKeyType prioritással rendelkezik! Egyenlőség esetén a lista első ilyen elemét.
   * Ha üres a lista, akkor None-t, egyébként Some((prioritás, érték))-et adjon vissza.
   */
  def minEntry(list: List[(MyKeyType,String)]):Option[(MyKeyType,String)] = ???


  /**
   * Írjunk teszteket, amik azt ellenőrzik, hogy random entry sorozat beszúrása után egy min ill. egy minEntry
   * művelet eredménye tényleg egy jó prioritást / entryt ad-e vissza!
   * Fussanak százszor.
   */
  "min" should "give the minimum priority after a series of inserts" in ???
  "minEntry" should "give the oldest minimum-priority entry after a series of inserts" in ???

  /**
   * Írjunk tesztet, ami azt ellenőrzik, hogy ha egy random entry listet generálunk, beszúrjuk az
   * elemeit, akkor a toList metódus jó sorrendben jó értékeket ad-e vissza! (Ehhez lehet, hogy nem árt
   * rendeztetnünk a random generált listánkat, van rá sortBy metódus.)
   */
  "toList" should "generate the values in increasing order of priorities" in ???

  /**
   * Írjunk tesztet, ami azt ellenőrzi, hogy egy random entry list generálása után, beszúrva az elemeit
   * egy üres prisorba, pontosan annyi pop után lesz-e üres a prisor, mint ahány elemet beszúrtunk!
   * Ezt is százszor.
   */
  "push-pop" should "pop exactly that many elements as many were inserted" in ???

  /**
   * Írjunk tesztet, ami egy hozzáadás - min - pop sorozatot generál és tesztel! Tartsunk nyilván egy
   * prioritás szerint rendezett listát és ellenőrizzük, hogy minden lépés után ha a kontroll listánk
   * pont akkor üres-e, amikor a prisorunk és ha nemüres, akkor ugyanaz-e a minimális prioritás mindkét
   * struktúrában. Majd random döntsünk, hogy pusholunk-e vagy (ha van mit) popolunk.
   */
  "push-pop-mix" should "behave exactly the way as a naive implementation" in ???

  /**
   * Ha minden teszt kizöldült, akkor gondolkodjunk el rajta, hogy ko- vagy kontravariánssá tudjuk-e tenni
   * a prioritát vagy az értéket az eredeti traitben és ha igen, akkor végezzük el a szükséges módosításokat!
   * Ha sikerül minden tesztet továbbra is kizöldíteni és az üres prisor egy case object lesz, akkor jó.
   */

}
